#include "internal.hpp"

namespace CaDiCaL {

/*------------------------------------------------------------------------*/

inline void
Internal::covered_literal_addition (int lit, Coveror & coveror) {
  require (COVER);
  assert (level == 1);
  LOG ("covered literal addition %d", lit);
  assert (!vals[lit]), assert (!vals[-lit]);
  vals[lit] = -1, vals[-lit] = 1;
  coveror.covered.push_back (lit);
  coveror.added.push_back (lit);
}

inline void
Internal::asymmetric_literal_addition (int lit, Coveror & coveror) {
  require (COVER);
  assert (level == 1);
  LOG ("asymmetric literal addition %d", lit);
  assert (!vals[lit]), assert (!vals[-lit]);
  vals[lit] = -1, vals[-lit] = 1;
  coveror.added.push_back (lit);
}

/*------------------------------------------------------------------------*/

// In essence propagation but with full occurrence lists.

inline bool
Internal::cover_propagate_asymmetric (int lit,
                                      Clause * ignore, Coveror & coveror)
{
  stats.propagations.cover++;
  require (COVER);
  assert (val (lit) < 0);
  bool subsumed = false;
  LOG ("asymmetric propagation of %d", lit);
  for (const auto c : occs (lit)) {
    if (c == ignore) continue;
    if (c->garbage) continue;
    int unit = 0;
    for (const auto & other : *c) {
      if (other == lit) continue;
      int tmp = val (other);
      if (tmp < 0) continue;
      if (tmp > 0 || unit) { unit = INT_MIN; break; }
      unit = other;
    }
    if (unit == INT_MIN) continue;
    if (!unit) {
      LOG (c, "subsuming");
      subsumed = true;
      break;
    }
    asymmetric_literal_addition (-unit, coveror);
    LOG (c, "asymmetric reason");
  }
  return subsumed;
}

// Covered literal addition (which needs full occurrence lists).

inline void
Internal::cover_propagate_covered (int lit, Coveror & coveror)
{
  require (COVER);
  assert (val (lit) < 0);
  if (frozen (lit)) {
    LOG ("no covered propagating on frozen literal %d", lit);
    return;
  }
  vector<int> intersection;
  LOG ("covered propagation of %d", lit);
  bool first = true;
  for (const auto d : occs (-lit)) {
    if (d->garbage) continue;
    bool blocked = false;
    for (const auto & other : *d) {
      if (other == -lit) continue;
      int tmp = val (other);
      if (tmp < 0) continue;
      if (tmp > 0) { blocked = true; break; }
      if (first) {
        intersection.push_back (other);
        mark (other);
      } else {
        tmp = marked (other);
        if (tmp > 0) unmark (other);
      }
    }
    if (blocked) {
      LOG (d, "blocked");
      unmark (intersection);
      intersection.clear ();
      continue;
    }
    if (!first) {
      const auto end = intersection.end ();
      auto j = intersection.begin ();
      for (auto i = j; i != end; i++) {
        const int other = *j++ = *i;
        const int tmp = marked (other);
        assert (tmp >= 0);
        if (tmp) j--, unmark (other);
        else mark (other);
      }
      intersection.resize (j - intersection.begin ());
    }
    first = false;
    if (intersection.empty ()) break;
  }
  if (intersection.empty ()) {
    LOG ("empty intersection of resolution candidate literals");
  } else {
    LOG (intersection,
      "non-empty intersection of resolution candidate literals");
    for (auto & lit : intersection) {
      covered_literal_addition (lit, coveror);
      unmark (lit);
    }
    intersection.clear ();
  }
}

bool Internal::cover_blocked (Clause * c) {
  require (COVER);
  assert (!c->garbage);
  LOG (c, "trying to cover by blocking");
  for (const auto & lit : *c) {
    if (frozen (lit)) continue;
    assert (val (lit) < 0);
    if (!var (lit).level) continue;
    LOG ("resolving on literal %d", lit);
    bool tautological = true;
    for (const auto & d : occs (-lit)) {
      if (d->garbage) continue;
      tautological = false;
      for (const auto & other : * d) {
        if (other == -lit) continue;
        int tmp = val (other);
        if (tmp <= 0) continue;
        LOG (d,
          "resolution on literal %d blocked by %d in",
          lit, other);
        tautological = true;
        break;
      }
      if (!tautological) {
        LOG (d, "not cover blocked resolving on literal %d", lit);
        break;
      }
    }
    if (tautological) {
      LOG (c, "all resolutions on %d cover blocked", lit);
      stats.cover.blocked++;
      stats.cover.total++;
      return true;
    }
  }
  LOG (c, "failed to cover block");
  return false;
}

bool Internal::cover_clause (Clause * c, Coveror & coveror) {

  require (COVER);
  assert (!c->garbage);

  LOG (c, "trying covered clauses elimination on");
  bool satisfied = false;
  for (const auto & lit : *c)
    if (val (lit) > 0)
      satisfied = true;

  if (satisfied) {
    LOG (c, "clause already satisfied");
    mark_garbage (c);
    return false;
  }

  assert (coveror.added.empty ());
  assert (coveror.covered.empty ());

  assert (!level);
  level = 1;
  LOG ("assuming literals of candidate clause");
  for (const auto & lit : *c)
    if (!val (lit))
      asymmetric_literal_addition (lit, coveror);

  bool tautological = false;
  coveror.next.asymmetric = coveror.next.covered = 0;

  while (!tautological) {
    if (coveror.next.asymmetric < coveror.added.size ()) {
      while (!tautological &&
        coveror.next.asymmetric < coveror.added.size ()) {
        int lit = coveror.added[coveror.next.asymmetric++];
        tautological = cover_propagate_asymmetric (lit, c, coveror);
      }
    } else if (coveror.next.covered < coveror.added.size ()) {
      int lit = coveror.added[coveror.next.covered++];
      cover_propagate_covered (lit, coveror);
    } else break;
  }

  if (!tautological && !c->garbage) tautological = cover_blocked (c);

  if (tautological) {
    if (coveror.covered.empty ()) {
      stats.cover.asymmetric++;
      stats.cover.total++;
      LOG (c, "asymmetric tautological");
      mark_garbage (c);
    } else {
      stats.cover.covered++;
      stats.cover.total++;
      LOG (c, "covered tautological");
      mark_garbage (c);
    }
  }

  assert (level == 1);
  for (const auto & lit : coveror.added)
    vals[lit] = vals[-lit] = 0;
  level = 0;

  coveror.covered.clear ();
  coveror.added.clear ();

  return tautological;
}

/*------------------------------------------------------------------------*/

bool Internal::cover () {

  return false;         // TODO get witnesses working then enable!

  if (!opts.cover) return false;
  if (unsat || terminating () || !stats.current.irredundant) return false;

  START_SIMPLIFIER (cover, COVER);

  stats.cover.count++;

  long delta = stats.propagations.search;
  delta -= last.cover.propagations;
  delta *= 1e-3 * opts.coverreleff;
  if (delta < opts.covermineff) delta = opts.covermineff;
  if (delta > opts.covermaxeff) delta = opts.covermaxeff;
  delta = max (delta, 2l* active ());

  PHASE ("cover", stats.cover.count,
    "covered clause elimination limit of %ld propagations", delta);

  long limit = stats.propagations.cover + delta;

  if (propagated < trail.size ()) {
    init_watches ();
    connect_watches ();
    LOG ("elimination produced %ld units", trail.size () - propagated);
    if (!propagate ()) {
      LOG ("propagating units before covered clause elimination "
        "results in empty clause");
      learn_empty_clause ();
    }
    reset_watches ();
    if (unsat) return false;
  }
  assert (!unsat);
  assert (propagated == trail.size ());

  init_occs ();

  vector<Clause *> schedule;
  Coveror coveror;

  for (int round = 0; round <= 1 && schedule.empty (); round++) {
    for (auto c : clauses) {
      if (c->garbage) continue;
      if (c->redundant) continue;
      bool satisfied = false;
      for (const auto & lit : *c)
        if (val (lit) > 0) { satisfied = true; break; }
      if (satisfied) { mark_garbage (c); continue; }
      for (const auto & lit : *c)
        occs (lit).push_back (c);
      if (c->covered) continue;
      schedule.push_back (c);
    }
  }

  stable_sort (schedule.begin (), schedule.end (),
    clause_smaller_size ());

  PHASE ("cover", stats.cover.count,
    "scheduled %zd clauses %.0f%% for covered clause elimination",
    schedule.size (),
    percent (schedule.size (), stats.current.irredundant));

  long covered = 0;
  while (!terminating () &&
         !schedule.empty () &&
         stats.propagations.cover < limit) {
    Clause * c = schedule.back ();
    schedule.pop_back ();
    assert (!c->covered);
    c->covered = true;
    if (cover_clause (c, coveror)) covered++;
  }

  reset_occs ();

  PHASE ("cover", stats.cover.count,
    "eliminated %ld covered clauses", covered);

  last.cover.propagations = stats.propagations.search;

  STOP_SIMPLIFIER (cover, COVER);

  report ('c', !opts.reportall && !covered);

  return covered;
}

}
